package unsw891.sml;

import java.util.Vector;
import unsw891.util.Debug;


/**
 * Internal Java structure representation of a SML document
 * <MODFIFICATION LOG>
 * Saturday 5PM - John - added clone method
 * Saturday 4pm - John - made toString() more efficient
 * Saturday 3pm - Stephen - created a new constructor, for colours
 * Friday 6PM - Harvey - added tags TAG_INVALID and TAG_ROOT constants
 * Friday 4PM - John - created
 */

public class Document {

	/**
	 * TAG_INVALID is used if a document has an invalid tag
	 * TAG_ROOT is used for the root document.
	 * Note: Maintain the relative order of these tags.
	 */
	public static final int TAG_INVALID = -1;
	public static final int TAG_ROOT = 0;
	public static final int TAG_TEXT = 1;
	public static final int TAG_B  =  2;
	public static final int TAG_EM =  3;
	public static final int TAG_I  =  4;
	public static final int TAG_PL =  5;
	public static final int TAG_S  =  6;
	public static final int TAG_TT =  7;
	public static final int TAG_U  =  8;
	public static final int TAG_0  =  9;
	public static final int TAG_1  = 10;
	public static final int TAG_2  = 11;
	public static final int TAG_3  = 12;
	public static final int TAG_4  = 13;
	public static final int TAG_5  = 14;
	public static final int TAG_6  = 15;
	public static final int TAG_7  = 16;
	public static final int TAG_8  = 17;
	public static final int TAG_9  = 18;
	public static final int TAG_r  = 19;
	public static final int TAG_g  = 20;
	public static final int TAG_b  = 21;
	public static final int TAG_c  = 22;
	public static final int TAG_m  = 23;
	public static final int TAG_y  = 24;
	public static final int TAG_k  = 25;
	public static final int TAG_w  = 26;

	private static final String[] OPEN_TAG = {    "",    "",  "<B>", "<EM>", "<I>",
	                                          "<PL>", "<S>", "<TT>", "<U>",  "<0>",
	                                           "<1>", "<2>",  "<3>", "<4>",  "<5>",
	                                           "<6>", "<7>",  "<8>", "<9>",  "<r>",
	                                           "<g>", "<b>",  "<c>", "<m>",  "<y>",
	                                           "<k>", "<w>"};

	private static final String[] CLOSE_TAG = {     "",     "",  "</B>", "</EM>", "</I>",
	                                           "</PL>", "</S>", "</TT>", "</U>",  "</0>",
	                                            "</1>", "</2>",  "</3>", "</4>",  "</5>",
	                                            "</6>", "</7>",  "</8>", "</9>",  "</r>",
	                                            "</g>", "</b>",  "</c>", "</m>",  "</y>",
	                                            "</k>", "</w>"};

	public  static final int[] TAG_SIZE = { 0, 0, 7, 9, 7,
	                                        9, 7, 9, 7, 7,
	                                        7, 7, 7, 7, 7,
	                                        7, 7, 7, 7, 7,
	                                        7, 7, 7, 7, 7,
	                                        7, 7};


	/**
	 * The type of tag this document belongs to. Must be one
	 * of the above.
	 */
	public int type = TAG_INVALID;

	/**
	 * Valid only if "type==TAG_TEXT"
	 */
	public String text;

	/**
	 * A link to its parent
	 * Generated only if the genLinks() function had been run. After a tree
	 * restructuring, you can either adjust the links yourself, or rerun
	 * genLinks() function, though less efficient (but safer).
	 */
	public Document parent;
	
	/**
	 * A vector of Document objects that are the children of this node
	 */
	public Vector children;

	
	/**
	 * Private constructor
	 */
	private Document() {
	}
	
	/**
	 *Create a Document object. Type must be in the range and is not TAG_TEXT
	 *or else a Document of type TAG_INVALID will be created
	 * <MODIFICATION LOG>
	 * Friday 8:50PM - John - Implemented
	 */
	public Document (int type) {
		if (type<TAG_INVALID || type>TAG_w || type==TAG_TEXT) {
			this.type = TAG_INVALID;
		} else {
			this.type = type;
		}

		children = new Vector();
	}


        /**
         *Create a Document object of a certain colour.
         */
        // by Stephen, 3pm Saturday
        public Document (char colour) {
          children = new Vector();
          switch(colour) {
            case 'r': type = TAG_r; break;
            case 'g': type = TAG_g; break;
            case 'b': type = TAG_b; break;
            case 'c': type = TAG_c; break;
            case 'm': type = TAG_m; break;
            case 'y': type = TAG_y; break;
            case 'k': type = TAG_k; break;
            case 'w': type = TAG_w; break;
            default : type = TAG_INVALID; break;
          }
        }

	/**
	 *Create a Document object of type TAG_TEXT, storing the given text string.
	 * <MODIFICATION LOG>
	 * Friday 8:50PM - John - Implemented
	 */
	public Document (int typeText, String text) {
		if (typeText==TAG_TEXT) {
			this.type = TAG_TEXT;
			this.text = text;
		} else {
			this.type = TAG_INVALID;
		}

		children = new Vector();
	}
	
	/**
	 *Generate links to the parent for all of its parent
	 */
	public void genLinks() {
		for (int i=0; i<children.size(); i++) {
			Document child = (Document)children.get(i);
			child.parent = this;
			child.genLinks();
		}
	}

	
	/**
	 *Convert this internal representation to an actual SML output
	 *<MODIFICATION LOG>
	 *Saturday 4pm - John - Converted it to use a StringBuffer for intermediate calculations
	 *Friday 9:42PM - John - Implemented it
	 */
	public String toString() {
		return toStringBuffer().toString();
	}
	
	/**
	 *Convert this internal representation to an actual SML output
	 *<MODIFICATION LOG>
	 *Saturday 4pm - John - Converted it to return a StringBuffer
	 *Friday 9:42PM - John - Implemented it
	 */
	public StringBuffer toStringBuffer() {
		if (type==TAG_INVALID) {
			System.err.println("Document.toStringBuffer() encountered an invalid tag");
			return new StringBuffer("*** INVALID TAG ***");
		}

		//print open tag <TAG>
		StringBuffer output = new StringBuffer(OPEN_TAG[type]);

		//print text if necessary
		if (type==TAG_TEXT) {
			output.append(text);
		}

		//print children
		for (int i=0; i<children.size(); i++) {
			output.append(((Document)(children.get(i))).toStringBuffer());
		}

		output.append(CLOSE_TAG[type]);

		return output;
	}

	/**
	 *<MODIFICATION LOG>
	 *Friday 10:09PM - John - Done efficient solution
	 *Friday 9PM - John - A simple inefficient implementation. Better to calculate
	 *the size directly rather than finding the length of the toString() string, since
	 *it wastes a lot of memory creating and destroying intermediate strings
	 */
	public int size() {
		if (type==TAG_INVALID) {
			System.err.println("Document.size() encountered an invalid tag");
			return Integer.MAX_VALUE;
		}

		//include cost of both opening and closing tags
		int size = TAG_SIZE[type];

		//include text cost if necessary
		if (type==TAG_TEXT) {
			size += text.length();
		}

		//add cost of children
		for (int i=0; i<children.size(); i++) {
			Document doc = (Document)(children.get(i));
			size += doc.size();
		}

		return size;
	}

  	/**
	 *<MODIFICATION LOG>
	 *Sunday 3pm - John - ignores text to give more accurate count
	 *Friday 10:09PM - John - Done efficient solution
	 *Friday 9PM - John - A simple inefficient implementation. Better to calculate
	 *the size directly rather than finding the length of the toString() string, since
	 *it wastes a lot of memory creating and destroying intermediate strings
	 */
	public int tagSize() {
		if (type==TAG_INVALID) {
			System.err.println("Document.size() encountered an invalid tag");
			return Integer.MAX_VALUE;
		}

		//include cost of both opening and closing tags
		int size = TAG_SIZE[type];


		//add cost of children
		for (int i=0; i<children.size(); i++) {
			Document doc = (Document)(children.get(i));
			size += doc.size();
		}

		return size;
	}


    /** Returns the type of this document, ie TAG_ROOT, TAG_TT etc.
        Convenience method only. Please do not remove.
        <MODIFICATION>
            1.0 by lawry
        */
    public int getType() {
        return type;
    }
    
    
    /** Returns the text stored in this document only if type is TAG_TEXT.
        Convenience method only. Please do not remove.
        <MODIFICATION>
            1.0 by lawry
        */
    public String getText() {
        Debug.assert(type == TAG_TEXT,
                "Document::getText() - Type is not TAG_TEXT");
        return text;
    }
	
	public Object clone() {
		Document doc = new Document();
		doc.type = type;
		doc.text = text;
		doc.children = new Vector();
		
		for (int i=0; i<children.size(); i++) {
			doc.children.add(((Document)(children.get(i))).clone());
		}
		
		return doc;
	}
	
}
