package unsw891.sml;

// created by Stephen!
// stores a part of the SML output, that is
//  * text, plus
//  * attributes, context, whatever.


public class Shard {

  public String text;
  public char   colour;
  public int    size;
  public int    flags;
  
  public static final int BOLD   = 0x40;
  public static final int ITALIC = 0x20;
  public static final int TT     = 0x10;
  public static final int STRONG = 0x08;
  public static final int EM     = 0x04;
  public static final int UNDER  = 0x03;
  
  public static final char UNINITIALISED_COLOUR = 'x';
  public static final int  UNINITIALISED_SIZE = -1;
  
  public boolean isBold() {
    return (flags & BOLD) != 0;
  }
  
  public void setBold(boolean x) {
    if (x) flags |= BOLD;
    else   flags &= 0xff ^ BOLD;
  }

  public boolean isItalic() {
    return (flags & ITALIC) != 0;
  }
  
  public void setItalic(boolean x) {
    if (x) flags |= ITALIC;
    else   flags &= 0xff ^ ITALIC;
  }

  public boolean isTT() {
    return (flags & TT) != 0;
  }
  
  public void setTT(boolean x) {
    if (x) flags |= TT;
    else   flags &= 0xff ^ TT;
  }

  public boolean isStrong() {
    return (flags & STRONG) != 0;
  }
  
  public void setStrong(boolean x) {
    flags &= 0xff ^ EM;  // automatically turns off EM
    if (x) flags |= STRONG;
    else   flags &= 0xff ^ STRONG;
  }

  public boolean isEm() {
    return (flags & EM) != 0;
  }
  
  public void setEm(boolean x) {
    if (x) flags |= EM;
    else   flags &= 0xff ^ EM;
  }

  public int isUnderline() {
    return (flags & UNDER);
  }
  
  public void setUnderline(int x) {
    flags &= 0xff ^ UNDER;
    flags |= x & UNDER;
  }

  public Shard() {
    text = null;
    flags = 0;
    colour = UNINITIALISED_COLOUR;
    size   = UNINITIALISED_SIZE;
  }

  public Shard(Shard old) {
    text = old.text;
    flags = old.flags;
    size = old.size;
    colour = old.colour;
  }
  
  public Shard(String newText, Shard old) {
    flags = old.flags;
    size = old.size;
    colour = old.colour;
    text = newText;
  }
  
  public void modify(int tagType) {
    switch(tagType) {
      case Document.TAG_B:  flags |= BOLD; break;
      case Document.TAG_EM:
        if ((flags & STRONG) == 0) flags ^= EM;
        break;
      case Document.TAG_I:  flags |= ITALIC; break;
      case Document.TAG_PL: flags = 0; break;
      case Document.TAG_S:  flags |= STRONG; break;
      case Document.TAG_TT: flags |= TT; break;
      case Document.TAG_U:
        if ((flags & UNDER) != 3) flags ++;
        break;
      case Document.TAG_r:  colour = 'r'; break; 
      case Document.TAG_g:  colour = 'g'; break;
      case Document.TAG_b:  colour = 'b'; break;
      case Document.TAG_c:  colour = 'c'; break;
      case Document.TAG_m:  colour = 'm'; break;
      case Document.TAG_y:  colour = 'y'; break;
      case Document.TAG_k:  colour = 'k'; break;
      case Document.TAG_w:  colour = 'w'; break;
      default: break;
    }
    if (tagType >= Document.TAG_0 && tagType <= Document.TAG_9) {
      size = tagType - Document.TAG_0;
    }
  }
  

  //********************************************************//
  // COMPARISON METHODS FOR SHARDS                          //
  //********************************************************//

  /**
   * Text equals will show if a context is equal to another one
   * for the purposes of validation
   */

  public boolean textEquals(Object foo) {	
	  if(!(foo instanceof Shard))
				  return false;

     
	  Shard o = (Shard) foo;

	  if((o.size != this.size) || (o.colour!= this.colour))
		  return false;		


	  if(!o.isStrong())
		  return flags == o.flags;
	  else
		  return (o.flags&( 0xff ^ EM))==(this.flags&( 0xff ^ EM));
  }

  /** 
   *	Whitespace has different contexts conditions, due to the fact
   * that it has certain attributes masked off. This function is to 
   * take into account that for whitespace markings, these 
   * context markers do not count.
   */
  public boolean whiteSpaceEquals(Object foo) {

	  if(!(foo instanceof Shard))
		  return false;

	  Shard o = (Shard) foo;

	  if((o.size != this.size))
		  return false;		

	  if(o.isUnderline()!=0)
		  if(o.colour != this.colour)
			  return false;

	  int mask = TT | UNDER;

	  return ((o.flags & mask) == (this.flags & mask)) ;	
  }
  
  
  /**
   * tells you if the first shard can be turned
   * into the second shard, by using colour tags, size tags
   * or B,I,TT,EM,S,U tags. <p> 
   * NOTE this disregards the existence of the PL tag <p>
   * @return (-1) if the conversion is impossible. Otherwise
   * returns an int indicating which properties need to be changed.
   * The bit fields used by Shard indicate which of the flags
   * need to be changed. The second least significant byte has
   * the form 000000sc where s is set if the size must be changed,
   * and c is set if the colour must be changed.
   */
  public int compare(Shard other) {
    int answer = 0;
    
    if (other.colour != this.colour) {
      if (other.colour == UNINITIALISED_COLOUR) return -1;
      else answer |= 0x100;
    }
    if (other.size != this.size) {
      if (other.size == UNINITIALISED_SIZE) return -1;
      else answer |= 0x200;
    }
    // otherwise we're right for colour and size conversion.
  
    if ((this.flags & UNDER) > (other.flags & UNDER)) return -1;
    else answer |= ((other.flags & UNDER) - (this.flags & UNDER));
    // that takes care of underlining
    
    // check for em-strong interaction
    int otherLocal = other.flags;
    if ((otherLocal & STRONG) > 0 && (this.flags & EM) > 0) {
      otherLocal |= EM;
    }
    
    // now check all the flags
    if (((this.flags | otherLocal) & 0xf8) != (otherLocal & 0xf8)) {
      return -1;
    }
    answer |= (this.flags ^ otherLocal) & 0xfc;
        
    return answer;
    
  } // end of compare
  
  
  public int changeCost(Shard other) {
    int answer = 0;
    
    if (other.colour != this.colour) {
      if (other.colour == UNINITIALISED_COLOUR) return 0xffffff;
      else answer += 7;
    }
    if (other.size != this.size) {
      if (other.size == UNINITIALISED_SIZE) return 0xffffff;
      else answer += 7;
    }
    // otherwise we're right for colour and size conversion.
  
    if ((this.flags & UNDER) > (other.flags & UNDER)) return 0xffffff;
    else answer += 7*((other.flags & UNDER) - (this.flags & UNDER));
    // that takes care of underlining
    
    // check for em-strong interaction
    int otherLocal = other.flags;
    if ((otherLocal & STRONG) > 0 && (this.flags & EM) > 0) {
      otherLocal |= EM;
    }
    
    // now check all the flags
	
    if (((this.flags | otherLocal) & 0x78) != (otherLocal & 0x78)) {
	  return 0xffffff;
	}
	
	if ((this.flags & BOLD) < (other.flags & BOLD))     answer += 7;
	if ((this.flags & ITALIC) < (other.flags & ITALIC)) answer += 7;
	if ((this.flags & TT) < (other.flags & TT))         answer += 9;
	if ((this.flags & STRONG) < (other.flags & STRONG)) answer += 7;
	if ((this.flags & EM) < (other.flags & EM))         answer += 9;
    
    return answer;
    
  
  } // end of changeCost
  

    /** This static method returns the root context: ie u = 0,
        b = em = i = s = tt = false, size = -1 color = 'x'.
        [MODIFICATION]
        1.0 - lawry - added by me first.
        */
    public static Shard getRootShard() {
        Shard s = new Shard();
        s.setBold(false);
        s.setEm(false);
        s.setItalic(false);
        s.setStrong(false);
        s.setTT(false);
        s.setUnderline(0);
        s.size = -1;
        s.colour = 'x';
        return s;
    }

	/* toString helper function - jordan, sat night */
	public String toString() {
		return "Shard Contains test: \"" + text + "\",size " + size + ",colour " + colour + " and flags " + flags; 
	}


    /** Returns true if the given shard is the same as this
        shard.
        */
    public boolean equals(Shard sh) {
        return (flags == sh.flags) &&
              (size == sh.size) &&
              (colour == sh.colour) &&
              (text.equals(sh.text));
    }


}
